///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      FILECHNG.CPP
//
//  Description
//      This file implements the directory watching function to trap event
//      on modification to the directory the service is watching.
//
//  Author
//      Irving De la Cruz (irvingd@microsoft.com)
//
//  Revision: 1.1
//
// Written for Microsoft Corporation.
// Copyright (c) 1996-1997 Microsoft Corporation. All rights reserved.
//
#include "DIRMON.H"

#define MAX_BUFFER  4096
#define DIRECTORY_WATCH_FLAGS (FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_LAST_WRITE)

typedef struct _DIRECTORY_INFO
{
    HANDLE      hDirectory;
    char        szDirectory[MAX_PATH];
    BYTE        Buffer[MAX_BUFFER];
    DWORD       dwBufLength;
    OVERLAPPED  Overlapped;
} DIRECTORY_INFO, *PDIRECTORY_INFO;

DIRECTORY_INFO  g_DirInfo = { 0 };
HANDLE g_hCompletionPort = NULL;
HANDLE g_hThread = NULL;

///////////////////////////////////////////////////////////////////////////////
//    HandleDirectoryChanges()
//
//    Parameters
//      [IN]    hCompletionPort   Handle for completion port
//
//    Purpose
//      This function receives notification of directory changes and 
//      calls PostPFMessage () to post a message in the Exchange PF of the
//      changes noticed. After the notification is processed this function
//      calls  ReadDirectoryChangesW again to reestablish the watch.
//
//    Return Value
//      A Win32 Error code.
//
DWORD WINAPI HandleDirectoryChange (HANDLE hCompletionPort)
{
    DWORD dwError = NO_ERROR, dwBytes, cbOffset;
    PDIRECTORY_INFO pDI;
    LPOVERLAPPED lpOverlapped;
    PFILE_NOTIFY_INFORMATION pFNI;
    do
    {
        // Retrieve the directory info for this directory
        // through the completion key
        GetQueuedCompletionStatus (hCompletionPort, 
                                   &dwBytes, 
                                   (LPDWORD)&pDI,
                                   &lpOverlapped,
                                   INFINITE);

        if (pDI)
        {
            pFNI = (PFILE_NOTIFY_INFORMATION)pDI->Buffer;
            do
            {
                cbOffset = pFNI->NextEntryOffset;
                PostPFMessage (pFNI);
                pFNI = (PFILE_NOTIFY_INFORMATION)((LPBYTE)pFNI + cbOffset);
            } while (cbOffset);
            ReadDirectoryChangesW (pDI->hDirectory,
                                   pDI->Buffer,
                                   MAX_BUFFER,
                                   FALSE,
                                   DIRECTORY_WATCH_FLAGS,
                                   &pDI->dwBufLength,
                                   &pDI->Overlapped,
                                   NULL);
        }
    } while (pDI);
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    StartDirectoryWatch()
//
//    Parameters
//      None
//
//    Purpose
//      Set up the completion port and the watching thread that handle the
//      directory change notifications
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI StartDirectoryWatch()
{
    HRESULT hResult = S_OK;
    lstrcpy (g_DirInfo.szDirectory, g_Config.szMonitorDir);
    // Get a handle on the directory
    g_DirInfo.hDirectory = CreateFile (g_DirInfo.szDirectory,
                                       FILE_LIST_DIRECTORY,
                                       FILE_SHARE_READ | 
                                       FILE_SHARE_WRITE | 
                                       FILE_SHARE_DELETE,
                                       NULL,
                                       OPEN_EXISTING,
                                       FILE_FLAG_BACKUP_SEMANTICS | 
                                       FILE_FLAG_OVERLAPPED,
                                       NULL);
    if (INVALID_HANDLE_VALUE == g_DirInfo.hDirectory)
    {
        hResult = GetLastError();
        return hResult;
    }
    // Associated the directory handle with a completion port handle
    g_hCompletionPort = CreateIoCompletionPort (g_DirInfo.hDirectory,
                                                NULL,
                                                (DWORD)&g_DirInfo,
                                                0);

    // Tell the system to notify us of changes in the directory
    // refered by the directory handle we just got
    ReadDirectoryChangesW (g_DirInfo.hDirectory,
                           g_DirInfo.Buffer,
                           MAX_BUFFER,
                           FALSE,
                           DIRECTORY_WATCH_FLAGS,
                           &g_DirInfo.dwBufLength,
                           &g_DirInfo.Overlapped,
                           NULL);
    
    // Dispatch the worker thread that handles the directory change notifications
    DWORD dwThreadID;
    g_hThread = CreateThread (NULL, 
                              0, 
                              (LPTHREAD_START_ROUTINE)HandleDirectoryChange, 
                              g_hCompletionPort, 
                              0, 
                              &dwThreadID);
    LogEvent (FALSE, EVENTMSG_I_DIR_WATCH_STARTED, g_DirInfo.szDirectory);
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    FinishDirectoryWatch()
//
//    Parameters
//      None.
//
//    Purpose
//      Terminate the thread that is waiting on directory change events. To
//      finish the thread we post a status notification to the completion
//      port associated with the directory watch function.
//
//    Return Value
//      None.
//
void WINAPI FinishDirectoryWatch()
{
    // Post a null status to the completion port queue
    PostQueuedCompletionStatus (g_hCompletionPort, 0, 0, NULL);
    CloseHandle (g_hCompletionPort);
    g_hCompletionPort = NULL;
    // Wait for the Directory thread to finish before exiting
    WaitForSingleObject (g_hThread, INFINITE);
    // Close the object
    CloseHandle (g_hThread);
    CloseHandle (g_DirInfo.hDirectory);
    LogEvent (FALSE, EVENTMSG_I_DIR_WATCH_STOPPED);
}

// End of file for FILECHNG.CPP
