///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MSGPOST.CPP
//
//  Description
//      This file implements the functions to post the directory changes to
//      a previously open IMAPIFolder object opened on the Exchange Public
//      Folder store.
//
//  Author
//      Irving De la Cruz (irvingd@microsoft.com)
//
//  Revision: 1.1
//
// Written for Microsoft Corporation.
// Copyright (c) 1996-1997 Microsoft Corporation. All rights reserved.
//
#include "DIRMON.H"

#define INITGUID
#define USES_PS_PUBLIC_STRINGS
#include <INITGUID.H>
#include <MAPIGUID.H>

typedef enum MSGPOST_EVENT
{
    MSGPOST_INVALID         = 0,
    MSGPOST_FILE_CREATED,
    MSGPOST_FILE_DELETED,
    MSGPOST_FILE_RENAMED,
    MSGPOST_FILE_MODIFIED
} MSGPOST_EVENT;

///////////////////////////////////////////////////////////////////////////////
//    UnicodeToAnsi()
//
//    Parameters
//      [IN]        wcStr       Buffer with UNICODE string
//      [IN-OUT]    szStr       Caller-allocated buffer for the ANSI string
//      [IN]        cchAnsi     Max number of bytes the szStr buffer can hold
//      
//    Purpose
//      Translate a string from UNICODE to ANSI in a caller-supplied buffer.
//
//    Return Value
//      None.
//
HRESULT WINAPI UnicodeToAnsi (LPWSTR wcStr, LPSTR  szStr, ULONG  cchAnsi)
{
    // Do we have a string to duplicate?
    if (!szStr || !wcStr)
    {
        return E_INVALIDARG;
    }
    int iChars = WideCharToMultiByte (CP_ACP,
                                      0,
                                      wcStr,
                                      -1,
                                      szStr,
                                      cchAnsi,
                                      NULL,
                                      NULL);
    if (0 == iChars)
    {
        DWORD dwError = GetLastError();
        LogEvent (TRUE, EVENTMSG_E_FAILED_UNICODE_TO_ANSI, dwError);
        return HRESULT_FROM_WIN32(dwError);
    }
    return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//    PostPFMessage()
//
//    Parameters
//      [IN]    pFNI    Information about the file
//
//    Purpose
//      This function process the changes in the directory we are watching.
//      The changes are posted to the public folder directory for every one
//      to see.
//
//    Return Value
//      None.
//
void WINAPI PostPFMessage (PFILE_NOTIFY_INFORMATION pFNI)
{
    char szFileName[MAX_PATH] = { 0 };
    static char szLastFileName[MAX_PATH] = { 0 };
    static char szSenderName[64] = { 0 };
    static DWORD dwLastCode = 0;
    static BOOL fInitIdentity = TRUE;

    MSGPOST_EVENT PostType = MSGPOST_INVALID;
    HRESULT hResult;
    if (fInitIdentity)
    {
        LPSPropValue pMyName;
        LPMAILUSER pUserObj;
        ULONG ulObjectType;
        // Lets get the name we have assigned as out primary identity. This is
        // the name that will go out in the FROM line of messages
        hResult = g_pSession->QueryIdentity (&g_IdentityEID.cb,
                                             (LPENTRYID *)&g_IdentityEID.lpb);
        if (!hResult)
        {
            hResult = g_pSession->OpenEntry (g_IdentityEID.cb,
                                             (LPENTRYID)g_IdentityEID.lpb, 
                                             NULL,
                                             0,
                                             &ulObjectType,
                                             (LPUNKNOWN *)&pUserObj);
            if (SUCCEEDED(hResult))
            {
                hResult = HrGetOneProp (pUserObj, PR_DISPLAY_NAME, &pMyName);
                if (!hResult)
                {
                    lstrcpyn (szSenderName, pMyName->Value.lpszA, sizeof(szSenderName)-1);
                    MAPIFreeBuffer (pMyName);
                    fInitIdentity = FALSE;
                }
                pUserObj->Release();
            }
        }
    }
    UnicodeToAnsi (pFNI->FileName, szFileName, MAX_PATH);
    if (pFNI->Action == dwLastCode && !lstrcmpi (szLastFileName, szFileName))
    {
        return;
    }
    if (FILE_ACTION_ADDED == pFNI->Action)
    {
        LogEvent (FALSE, EVENTMSG_V_FILE_CREATED, szFileName);
        PostType = MSGPOST_FILE_CREATED;
    }
    if (FILE_ACTION_MODIFIED == pFNI->Action)
    {
        if (FILE_ACTION_ADDED == dwLastCode && !lstrcmpi (szLastFileName, szFileName))
        {
            pFNI->Action = 0;
            ZeroMemory (szFileName, sizeof(szFileName));
            // After a file is create the OS dispatches two notifs:
            // FILE_ACTION_ADDED immediately followed by FILE_ACTION_MODIFIED.
            // In this case we only want to post one message.
        }
        else
        {
            LogEvent (FALSE, EVENTMSG_V_FILE_MODIFIED, szFileName);
            PostType = MSGPOST_FILE_MODIFIED;
        }
    }
    if (FILE_ACTION_REMOVED == pFNI->Action)
    {
        LogEvent (FALSE, EVENTMSG_V_FILE_DELETED, szFileName);
        PostType = MSGPOST_FILE_DELETED;
    }
    if (FILE_ACTION_RENAMED_NEW_NAME == pFNI->Action)
    {
        LPSTR szStrs[] = 
        {
            szLastFileName,
            szFileName
        };
        LogEvent (FALSE, EVENTMSG_V_FILE_DELETED, 0, szStrs, 2);
        PostType = MSGPOST_FILE_RENAMED;
    }
    lstrcpy (szLastFileName, szFileName);
    dwLastCode = pFNI->Action;
    if (!PostType)
    {
        return;
    }
    // This change must be posted. Create a new IMessage object and set its properties
    LPMESSAGE pMessage;
    // Use MAPI_DEFERRED_ERRORS to improve perfomance
    hResult = g_pFolder->CreateMessage (NULL, MAPI_DEFERRED_ERRORS, &pMessage);
    if (hResult)
    {
        LogEvent (TRUE, EVENTMSG_E_FAILED_CREATEMESSAGE, hResult);
        return;
    }

    // Allocate an array of properties and fill them up. Them order is irrelevant.
    DWORD i = 0;
    #define MSGPROPS    20
    SPropValue spvMsgProps[MSGPROPS] = { 0 };
    char szBody[256] = { 0 }, szSubject[256] = { 0 };
    
    // The file name is the TOPIC so that all
    // transactions on a given file may be group
    spvMsgProps[i].ulPropTag   = PR_CONVERSATION_TOPIC;
    spvMsgProps[i].Value.lpszA = szFileName;
    i++;

    // This flag will cause the message to appear originally "UNREAD"
    spvMsgProps[i].ulPropTag   = PR_MESSAGE_FLAGS;
    spvMsgProps[i].Value.l     = MSGFLAG_UNMODIFIED;
    i++;

    // Set the appropiate body and subject base on the
    // actual notification of the directory change
    switch (PostType)
    {
        case MSGPOST_FILE_CREATED :
            wsprintf (szSubject, "FILE CREATED - %s", szFileName);
            wsprintf (szBody, "Notification of Change: FILE CREATED - File Name: %s", szFileName);
            break;
        case MSGPOST_FILE_DELETED :
            wsprintf (szSubject, "FILE DELETED - %s", szFileName);
            wsprintf (szBody, "Notification of Change: FILE DELETED - File Name: %s", szFileName);
            break;
        case MSGPOST_FILE_RENAMED :
            wsprintf (szSubject, "FILE RENAMED - %s", szFileName);
            wsprintf (szBody, "Notification of Change: FILE RENAMED - File Name: %s", szFileName);
            break;
        case MSGPOST_FILE_MODIFIED :
            wsprintf (szSubject, "FILE MODIFIED - %s", szFileName);
            wsprintf (szBody, "Notification of Change: FILE MODIFIED - File Name: %s", szFileName);
            break;
    }

    // The PR_SENDER_xx and PR_SENT_xx identify who posted the message.
    // Note that the entry ID of the PR_SENDER_xx and PR_SENT_xx
    // are the same. You could set them differently to implement a
    // "post on behalf of" functionality
    // On the entry ID and and the display name are need the other
    // identity properties (i.e. PR_SENDER_SEARCH_KEY,
    // PR_SENDER_EMAIL_ADDRESS, etc.) are not needed.
    spvMsgProps[i].ulPropTag    = PR_SENDER_ENTRYID;
    spvMsgProps[i].Value.bin    = g_IdentityEID;
    i++;

    spvMsgProps[i].ulPropTag    = PR_SENDER_NAME;
    spvMsgProps[i].Value.lpszA  = szSenderName;
    i++;

    spvMsgProps[i].ulPropTag    = PR_SENT_REPRESENTING_ENTRYID;
    spvMsgProps[i].Value.bin    = g_IdentityEID;
    i++;

    spvMsgProps[i].ulPropTag    = PR_SENT_REPRESENTING_NAME;
    spvMsgProps[i].Value.lpszA  = szSenderName;
    i++;

    // This is the message class of the new message
    spvMsgProps[i].ulPropTag    = PR_MESSAGE_CLASS;
    spvMsgProps[i].Value.lpszA  = "IPM.POST";
    i++;

    spvMsgProps[i].ulPropTag    = PR_PRIORITY;
	spvMsgProps[i].Value.l      = PRIO_NORMAL;
    i++;

	spvMsgProps[i].ulPropTag    = PR_SUBJECT;
	spvMsgProps[i].Value.lpszA  = szSubject;
    i++;
	
    spvMsgProps[i].ulPropTag    = PR_BODY;
	spvMsgProps[i].Value.lpszA  = szBody;
    i++;

    spvMsgProps[i].ulPropTag    = PR_CLIENT_SUBMIT_TIME;
    GetSystemTimeAsFileTime (&spvMsgProps[i].Value.ft);	
    i++;

    spvMsgProps[i].ulPropTag    = PR_CLIENT_SUBMIT_TIME;
	spvMsgProps[i].Value.ft     = spvMsgProps[i - 1].Value.ft;
    i++;

    spvMsgProps[i].ulPropTag    = PR_DELETE_AFTER_SUBMIT;
	spvMsgProps[i].Value.b      = TRUE;
    i++;

    // Set all proerties at once. Batching the setting of the
    // properties improves perfomance on provides that
    // transmit data over a network (i.e. Exchange)
    hResult = pMessage->SetProps (i, spvMsgProps, NULL);

    // To post a message simply call SaveChanges() after its properties are set.
    // Note that typically a posted message does not have a recipient table
    pMessage->SaveChanges (0);

    pMessage->Release();    
}

// End of file for MSGPOST.CPP
